<?php

declare(strict_types=1);

class ControllerExtensionPaymentConnectPay extends Controller {

  private array $error = [];
  private const REQUEST_URL = 'https://api2.connectpay.com/';

  public function index()
  {
    $this->load->model('checkout/order');
    $this->load->model('extension/payment/connectpay');
    $this->load->language('extension/payment/connectpay');

    if ($this->request->server['REQUEST_METHOD'] == 'POST') {
      if (!isset($this->session->data['order_id'])) {
        return false;
      }

      $orderData = $this->model_checkout_order->getOrder($this->session->data['order_id']);
      $paymentMethods = explode(' ', $this->request->post['payment_methods']);
      $paymentMethod = $paymentMethods[0];
      if (isset($paymentMethods[1])) {
        $paymentMethodId = $paymentMethods[1];
        $orderData['cp_payment_method_id'] = $paymentMethodId;
      }
      $orderData['cp_payment_method'] = [$paymentMethod];

      $orderData['noPaymentProvidersEnabled'] = 0;
      if (isset($this->request->post['noPaymentProvidersEnabled'])) {
        $orderData['noPaymentProvidersEnabled'] = $this->request->post['noPaymentProvidersEnabled'];
      }

      if (!$this->validate($orderData)) {
        foreach ($this->error as $error) {
          $this->session->data['error'] = $error;
        }
        $this->response->redirect($this->url->link('checkout/checkout'));
      }

      $response = json_decode($this->initiatePayment($orderData), true);
      if (isset($response['payment']['redirectUrl'])) {
        $this->model_extension_payment_connectpay->addTransaction($response['payment']['paymentId'], (int)$orderData['order_id']);
        $this->model_checkout_order->addOrderHistory($this->session->data['order_id'], $this->config->get('payment_connectpay_order_status_pending'));
        $this->response->redirect($response['payment']['redirectUrl']);
      }

      $this->response->redirect($this->url->link('checkout/failure'));
    }

    $data['action'] = $this->url->link('extension/payment/connectpay', '', true);
    $data['payment_providers'] = $this->config->get('payment_connectpay_payment_providers');
    $data['payment_methods'] = $this->config->get('payment_connectpay_payment_method');
    $data['payment_country'] = $this->config->get('payment_connectpay_payment_country');
    $data['all_available_countries'] = $this->config->get('payment_connectpay_all_countries');

    return $this->load->view('extension/payment/connectpay', $data);
  }

  public function cron(): void
  {
    $this->load->model('extension/payment/connectpay');
    $transactions = $this->model_extension_payment_connectpay->getTransactions();
    $firstCheckInterval = 4;

    foreach ($transactions as $transaction) {
      $transactionDate = new DateTime($transaction['date_added']);
      $dateTimeNow = new DateTime();
      $diffInHours = $transactionDate->diff($dateTimeNow)->h;

      if ($diffInHours >= $firstCheckInterval) {
        $this->updateOrderStatusByTransactionStatus($transaction);
      }
    }

    $this->log->write('ConnectPay payment status check executed...');
  }

  public function validate(array $orderData): bool
  {
    if (isset($orderData['firstname'])
      && strlen($orderData['firstname']) > 50
    ) {
        $this->error['firstname'] = $this->language->get('error_cp_firstname');
    }

    if (isset($orderData['lastname'])
      && strlen($orderData['lastname']) > 50
    ) {
      $this->error['lastname'] = $this->language->get('error_cp_lastname');
    }

    if (isset($orderData['email'])
      && strlen($orderData['email']) > 250
    ) {
      $this->error['email'] = $this->language->get('error_cp_email');
    }

    if (isset($orderData['telephone'])
      && (
        strlen($this->removeSpecialCharacters($orderData['telephone'])) > 15
        || !is_numeric($this->removeSpecialCharacters($orderData['telephone']))
      )
    ) {
      $this->error['telephone'] = $this->language->get('error_cp_phone');
    }

    return !$this->error;
  }

  public function callback(): void
  {
    $this->load->model('extension/payment/connectpay');
    $this->load->model('checkout/order');
    if (isset($this->request->get['orderId'])) {
      $orderId = $this->request->get['orderId'];
      $transaction = $this->model_extension_payment_connectpay->getTransactionByOrderId($orderId);
      if (!empty($transaction) && isset($transaction[0]['payment_id'])) {
        $paymentId = $transaction[0]['payment_id'];
        $transaction = json_decode($this->getPaymentDetails($paymentId), true);
        if (isset($transaction['status'])
          && isset($transaction['status']['code'])
        ) {
          $transactionStatus = $transaction['status']['code'];
          switch ($transactionStatus) {
            case 'Abandoned':
            case 'Failed':
              $this->response->redirect($this->url->link('checkout/failure'));
              break;
            default:
              $this->response->redirect($this->url->link('checkout/success'));
          }
        }
      } elseif ($this->model_checkout_order->getOrder($orderId)['order_status_id'] !== $this->config->get('payment_connectpay_order_status_failed_abandoned')) {
        $this->response->redirect($this->url->link('checkout/success'));
      } else {
        $this->response->redirect($this->url->link('checkout/failure'));
      }
    }
  }

  public function notification()
  {
    $this->load->model('extension/payment/connectpay');
    $this->load->model('checkout/order');
    // JSON body
    $data = json_decode(file_get_contents('php://input'), TRUE);
    // incoming notification logging
    $this->log->write(file_get_contents('php://input'));
    // remittanceInformationUnstructured = Order ID
    $orderId = $data['payment']['remittanceInformationUnstructured'];
    $transaction = $this->model_extension_payment_connectpay->getTransactionByOrderId($orderId);

    if (!empty($transaction) && isset($transaction[0]['payment_id'])) {
      $this->updateOrderStatusByTransactionStatus($transaction[0]);
    }
  }

  private function initiatePayment(array $orderData): string
  {
    $storeUrl = HTTPS_SERVER ?? HTTP_SERVER;
    $providerCountryCode = $this->config->get('payment_connectpay_payment_country');
    $userLanguage = strtoupper($this->language->get('code'));
    $requestUrl = self::REQUEST_URL . 'merchant/payments';
    $redirectUrl = $storeUrl . '?route=extension/payment/connectpay/callback?&orderId=' . $orderData['order_id'];
    $notificationUrl = $storeUrl . '?route=extension/payment/connectpay/notification';

    $brandId = $this->config->get('payment_connectpay_brand_id');
    $ip = $orderData['ip'];
    $reqMethod = strtolower('post');
    $requestUrlForSignature = strtolower($requestUrl);

    // Get URL without https:// part
    $requestUrlForSignature = explode('//', $requestUrlForSignature)[1];

    // Request body
    $requestBody = [
      'merchant' => [
        'brandId' => $brandId,
        'redirectUrl' => $redirectUrl,
        'notificationUrl' => $notificationUrl,
        'language' => $userLanguage,
      ],
      'payment' => [
        'paymentMethod' => $orderData['cp_payment_method'],
        'providerCountryCode' => $providerCountryCode,
        'instructedAmount' => [
          'amount' => number_format((float)$orderData['total'], 2),
          'currency' => $orderData['currency_code'],
        ],
        'remittanceInformationUnstructured' => $orderData['order_id']
      ],
      'identifiers' => [
        'merchantReference' => $this->getGuid(),
        'endToEndId' => $this->randomString(30),
      ],
      'consumer' => [
        'firstName' => $orderData['payment_firstname'],
        'lastName' => $orderData['payment_lastname'],
        'ipAddress' => $ip,
        'consumerId' => $this->getGuid(),
        'phoneNo' => $this->removeSpecialCharacters($orderData['telephone']),
        'email' => $orderData['email'],
      ]
    ];

    if ($providerCountryCode !== 'DE'
      && $this->config->get('payment_connectpay_all_countries') === '0'
      && $orderData['cp_payment_method'][0] != 'CARD'
      && $orderData['noPaymentProvidersEnabled'] == '0'
    ) {
      $requestBody['payment']['providerId'] = $orderData['cp_payment_method_id'];
    }

    if ($this->config->get('payment_connectpay_all_countries') === '1') {
      $requestBody['payment']['providerCountryCode'] = $this->config->get('payment_connectpay_default_payment_country');
    }

    $requestBodyJson = json_encode($requestBody);
    $requestBodyJson = preg_replace('/\s+/', '', $requestBodyJson);

    // Data to sign signature
    $dataToSign = $reqMethod . '|' . $requestUrlForSignature . '|' . $requestBodyJson;

    // Request headers
    $headers = [
      'Authorization: ' . $this->getAuthToken(),
      'X-Request-ID: ' . $this->getGuid(),
      'Accept: ' . 'application/json;version=2',
      'Content-Type: application/json',
      'X-Signature: ' . $this->getSignature($dataToSign)
    ];

    // Response from POST Initiate Payment request
    $curl = curl_init();
    curl_setopt_array($curl, [
      CURLOPT_URL => $requestUrl,
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'POST',
      CURLOPT_POSTFIELDS => $requestBodyJson,
      CURLOPT_HTTPHEADER => $headers
    ]);

    curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
    $response = curl_exec($curl);
    curl_close($curl);

    if (false === $response) {
      return '';
    }

    if (curl_getinfo($curl, CURLINFO_HTTP_CODE) !== 200
        || curl_getinfo($curl, CURLINFO_HTTP_CODE) !== 201
    ) {
      $responseBody = json_decode($response, true);
      foreach ($responseBody['messages'] ?? [] as $message) {
        $this->log->write($message);
      }
    }

    return $response;
  }

  private function getAuthToken(): string
  {
    $clientId = $this->config->get('payment_connectpay_client_id');
    $clientSecret = $this->config->get('payment_connectpay_client_secret');

    return 'Basic ' . base64_encode($clientId . ':' . $clientSecret);
  }

  private function getGuid(): string
  {
    mt_srand((int)microtime() * 10000);
    $charid = md5(uniqid((string)rand(), true));
    $hyphen = chr(45);

    return substr($charid, 0, 8) . $hyphen
      . substr($charid, 8, 4) . $hyphen
      . substr($charid, 12, 4) . $hyphen
      . substr($charid, 16, 4) . $hyphen
      . substr($charid, 20, 12);
  }

  private function getSignature(string $data): string
  {
    $privateKey = base64_decode($this->config->get('payment_connectpay_private_key'));
    $signature = '';
    $alg = 'sha256WithRSAEncryption';
    openssl_sign($data, $signature, $privateKey, $alg);

    return base64_encode($signature);
  }

  private function randomString($length)
  {
    // String of all alphanumeric character
    $randomString = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
    // Shuffle the $randomString and returns substring of specified length
    return substr(str_shuffle($randomString), 0, $length);
  }

  private function removeSpecialCharacters(string $string): string
  {
    // Only digits will be returned
    return preg_replace('/\D/', '', $string);
  }

  private function updateOrderStatusByTransactionStatus(array $paymentTransaction): void
  {
    $this->load->model('checkout/order');
    $this->load->model('localisation/order_status');
    $transaction = json_decode($this->getPaymentDetails($paymentTransaction['payment_id']), true);
    $orderId = $paymentTransaction['order_id'];

    if (isset($transaction['status'])
      && isset($transaction['status']['code'])
      && true === (bool)$transaction['status']['isTerminal']
    ) {
      $transactionStatus = $transaction['status']['code'];

      // Change order status by transaction code mapping in settings
      switch ($transactionStatus) {
        case 'Approved':
        case 'Settled':
          $this->model_checkout_order->addOrderHistory((int)$orderId, $this->config->get('payment_connectpay_order_status_approved_settled'));
          $this->model_extension_payment_connectpay->removeTransaction($paymentTransaction['payment_id']);
          break;
        case 'Abandoned':
        case 'Failed':
          $this->model_checkout_order->addOrderHistory((int)$orderId, $this->config->get('payment_connectpay_order_status_failed_abandoned'));
          $this->model_extension_payment_connectpay->removeTransaction($paymentTransaction['payment_id']);
          break;
      }
    }
  }

  private function getPaymentDetails(string $paymentId): string
  {
    $reqMethod = 'get';
    $requestUrl = self::REQUEST_URL . 'merchant/payments/' . $paymentId ;

    // Data to sign signature
    $dataToSign = $this->getDataToSign($requestUrl, $reqMethod);

    // Request headers
    $headers = [
      'Authorization: ' . $this->getAuthToken(),
      'X-Request-ID: ' . $this->getGuid(),
      'Accept: ' . 'application/json;version=2',
      'X-Signature: ' . $this->getSignature($dataToSign)
    ];

    $curl = curl_init();

    curl_setopt_array($curl, [
      CURLOPT_URL => $requestUrl,
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'GET',
      CURLOPT_HTTPHEADER => $headers
    ],
    );

    $response = curl_exec($curl);
    curl_close($curl);

    if (curl_getinfo($curl, CURLINFO_HTTP_CODE) !== 200
      || curl_getinfo($curl, CURLINFO_HTTP_CODE) !== 201
    ) {
      $responseBody = json_decode($response, true);
      foreach ($responseBody['messages'] ?? [] as $message) {
        $this->log->write($message);
      }
    }

    return $response;
  }

  private function getDataToSign(string $requestUrl, string $reqMethod): string
  {
    $requestUrlForSignature = strtolower($requestUrl);
    $requestMethod = strtolower($reqMethod);
    $requestBody = strtolower('');

    // Get URL without https:// part
    $requestUrlForSignature = explode('//', $requestUrlForSignature)[1];

    // Data to sign signature
    return $requestMethod . '|' . $requestUrlForSignature . '|' . $requestBody;
  }
}